/* Semaphore Manager:  */

#ifdef __USAGE
%C [options] &

Options:
   -s   semaphore name. Default is 'qnx/sema1'

e.g. 
%C -s qnx/sema2 &
#endif

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <sys/psinfo.h>
#include <sys/timers.h>
#include <sys/kernel.h>
#include <sys/proxy.h>
#include <sys/name.h>
#include <sys/sys_msg.h>
#include <sys/types.h>
#include <sys/osinfo.h>
#include <unistd.h>
#include "sema.h"

#define MAX_PIDS   	50    /* maximum number of processes that can use this semaphore */
#define SNAME_SIZE 	16    /* the maximum length of a sema4 name in bytes */

char sema4name[SNAME_SIZE+1];

struct sema4entry {
    pid_t  pid;              /* process who is in sema4 list */
    int    notify_method;    /* SEMA_REPLY_TO_ME or SEMA_SIGNAL_ME */	
    int    signal;           /* used if notify_method == SEMA_SIGNAL_ME */
    int    priority;         /* priority of semaphore request. 0 = lowest */
    double timestamp;        /* time when semaphore was requested */
    };
    
/*
 * All messages received have this structure
 */
struct msg_entry {
    signed char msg_type;
    int    notify_method;    /* SEMA_REPLY_TO_ME or SEMA_SIGNAL_ME */	
    int    signal;           /* used if notify_method == SEMA_SIGNAL_ME */
    int    priority;         /* priority of semaphore request. 0 = lowest */
	};

pid_t pid;	    				     /* Process id of last QNX msg */
struct sema_msg_entry msg;		     /* The message buffer */
struct sema4entry sema[MAX_PIDS];    /* The sema4 table */
int    num_pids = 0;                 /* number of processes in the table */
int    whose_got_it = 0;             /* pid who currently has semaphore. 0 = none */

volatile struct _timesel far *t_ptr;
volatile long s_sec, s_nsec;         /* sec is second part, nsec = nanoseconds */

int  add_to_list( void );
int  release_from_list( void );
int  release_sema4( void );
void sort_list( void );
void check_semaphore( void );

main(int argc, char **argv)
	{
	extern void sema_exit();
	int i;
    unsigned opt;
    struct {
        struct _sysmsg_hdr      hdr;
    } *deathmsg = &msg;          
    int pflags;

    strcpy(sema4name, "qnx/sema1");

	while((opt=getopt(argc,argv,"s:"))!=-1) {
		switch(opt) {
 			case 's': /* semaphore name */
                      strncpy(sema4name, optarg, SNAME_SIZE);
                      sema4name[SNAME_SIZE] = '\0';
                      break;
			default:  /* invalid option */
                      exit(EXIT_FAILURE);	
                      break;
		    }
	    }

	t_ptr = MK_FP(0xb0,0);     /* get a pointer into system time */

	/*
	 * Register a name so processes can send us messages.
	 */
	if((qnx_name_attach(0, sema4name)) == -1) {
		fprintf(stderr, "Unable to attach name %s: %s\n",sema4name,strerror(errno));
		exit(-1);
		}

	for(opt = _SIGMIN ; opt < _SIGMAX ; ++opt)
		signal(opt, &sema_exit);
	signal(SIGPWR, SIG_IGN);

    pflags = _PPF_INFORM | _PPF_PRIORITY_REC;
	if(qnx_pflags(~0L, pflags, 0, 0)) {
		fprintf(stderr, "\nUnable to become informed admin: %s.\n", strerror(errno));
		exit(-1);
		}

	setpgid(0, 0); /* be pgrp leader */
    close(0);      /* close up stdin */
    close(1);      /* close up stdout */
    close(2);      /* close up stderr */
    fcloseall();   /* to be safe ... close any open files */

    for ( i=0; i < MAX_PIDS; i++ ) {
        sema[i].pid = 0;
        sema[i].notify_method = 1;
        sema[i].signal = 0;
        sema[i].priority = -1;
        sema[i].timestamp = 0.0;
        }

	/*
	 * Now we block waiting for a message.
	 */
	for(;;) {
    getmessage:
	pid = Receive(0, &msg, sizeof(msg));

    /* look for death messages first */
    if ( deathmsg->hdr.type       == _SYSMSG
         && deathmsg->hdr.subtype == _SYSMSG_SUBTYPE_DEATH ) {
           Reply(pid, 0, 0);
           release_from_list();
           goto getmessage;
           }
    switch ( msg.msg_type ) {
			case SEMA_ADD_TO_LIST:
				msg.msg_type = add_to_list();
                if ( msg.msg_type != SEMA_OK )
                   Reply( pid, &msg, 1 );
                else
                   if ( msg.notify_method == SEMA_SIGNAL_ME )
                      Reply( pid, &msg, 1 );
				break;

			case SEMA_RELEASE_FROM_LIST:
				msg.msg_type = release_from_list();
                Reply( pid, &msg, 1 );
				break;

		    default:
			    msg.msg_type = SEMA_BAD_MSG;
			    Reply(pid, &msg, 1);
                break;
			}
    }
}

/*
 * Add me to semaphore list
 */
int add_to_list( void )
{
    if ( num_pids == MAX_PIDS - 1 )
       return(SEMA_NO_ENTRIES);

    num_pids++;
    sema[num_pids].notify_method = msg.notify_method;
    sema[num_pids].signal = msg.signal;
    sema[num_pids].pid = pid;
    sema[num_pids].priority = msg.priority;
    s_sec  = t_ptr->seconds;		
    s_nsec = t_ptr->nsec;		
    sema[num_pids].timestamp = (double)s_sec + ((double)s_nsec/1000000000.0);
    sort_list();
    check_semaphore();
	return ( SEMA_OK );	
}

void check_semaphore( void )
{
    if ( whose_got_it == 0 ) {   /* no one has it */
       if ( sema[0].pid != 0 ) {
          /* printf("%d\n",sema[0].pid); */
          switch ( sema[0].notify_method ) {  /* first one in list */
             case SEMA_REPLY_TO_ME: msg.msg_type = SEMA_OK;
                                    Reply( sema[0].pid, &msg, 1 ); 
                                    break;
             case SEMA_SIGNAL_ME:   kill( sema[0].pid, sema[0].signal );
                                    break;
             }
          whose_got_it = sema[0].pid;
          }
       }
}

/*
 * Remove me from semaphore list
 */
int release_from_list( void )
{
    int i;
    int num_finds=0;

    for ( i=0; i < num_pids; i++ ) {
        if ( pid == sema[i].pid ) {
           sema[i].pid = 0;
           sema[i].priority = -1;
           sema[i].signal = 0;
           sema[i].notify_method = 1;
           sema[i].timestamp = 0.0;
           num_finds++;
           }
        }

    if ( num_finds > 0 ) {
       num_pids -= num_finds;
       sort_list();
       if ( pid == whose_got_it ) {
          whose_got_it = 0;
          check_semaphore();
          }
       }
	return ( SEMA_OK );	
}

int compare( struct sema4entry *t1, struct sema4entry *t2 ) {
   if ( t1->priority > t2->priority )
      return(-1);
   else if ( t1->priority < t2->priority )
      return(1);
   else {
      /* now check time stamp, the priorities are the same. make sure the */
      /* earliest time gets to the top */
      if ( t1->timestamp < t2->timestamp )
         return(-1);
      else if ( t1->timestamp > t2->timestamp )
         return(1);
      else
         return(0);    
      }
}
void sort_list( void )
{
qsort( sema, num_pids+1, sizeof(struct sema4entry), compare );
}

void sema_exit( void )
    {
	pid_t pid;
	unsigned short msg = EOK;  /* must be this type for system messages */

	qnx_pflags(0, _PPF_INFORM, 0, 0);
	while((pid = Creceive(0, 0, 0)) != -1  &&  pid != 0)
		Reply(pid, &msg, sizeof(msg));
	exit(0);
	}

